#ifndef __PYX_HAVE__MACS2__IO__cParser
#define __PYX_HAVE__MACS2__IO__cParser


#ifndef __PYX_HAVE_API__MACS2__IO__cParser

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

__PYX_EXTERN_C DL_IMPORT(int) HAS_PYSAM;

#endif /* !__PYX_HAVE_API__MACS2__IO__cParser */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initcParser(void);
#else
PyMODINIT_FUNC PyInit_cParser(void);
#endif

#endif /* !__PYX_HAVE__MACS2__IO__cParser */
